# Display time-power-squared values for waveform data
from math import sqrt
from typing import List, Tuple, Union

import numpy as np
from PySide2 import QtWidgets, QtCore

from sohstationviewer.conf import constants as const
from sohstationviewer.controller.plotting_data import (
    get_title, get_day_ticks, format_time,
)
from sohstationviewer.controller.util import (
    display_tracking_info, add_thousand_separator,
)
from sohstationviewer.database.extract_data import (
    get_color_def, get_color_ranges, get_chan_label,
)
from sohstationviewer.model.data_type_model import DataTypeModel
from sohstationviewer.model.handling_data import (
    get_start_5mins_of_diff_days, find_tps_tm_idx,
)
from sohstationviewer.view.util.enums import LogType
from sohstationviewer.view.plotting.plotting_widget import plotting_widget
from sohstationviewer.view.plotting.time_power_squared_processor import (
    TimePowerSquaredProcessor,
)
from sohstationviewer.view.util.color import clr


class TimePowerSquaredWidget(plotting_widget.PlottingWidget):
    stopped = QtCore.Signal()
    """
    Widget to display time power square data for waveform channels
    """
    def __init__(self, *args, **kwarg):
        """
        rulers: [matplotlib.lines.Line2D,] - list of squares on each waveform
            channels to highlight the five-minute at the mouse click on
            TimePowerSquaredWidget or the five-minute corresponding to the
            time at the mouse click on other plotting widgets
        """
        self.rulers = []
        """
        zoom_marker1s: [matplotlib.lines.Line2D,] - list of line markers to
            mark the five-minute corresponding to the start of the zoom area
        """
        self.zoom_marker1s = []
        """
        zoom_marker2s: [matplotlib.lines.Line2D,] - list of line markers to
            mark the five-minute corresponding to the end of the zoom area
        """
        self.zoom_marker2s = []
        """
        every_day_5_min_list: [[288 of floats], ] - the list of all starts
            of five minutes for every day in which each day has 288 of
            5 minutes.
        """
        self.start_5mins_of_diff_days = []
        """
        tps_t: float - prompt's time on tps's chart to help rulers on other
            plotting widgets to identify their location
        """
        self.tps_t = 0

        self.tps_processors: List[TimePowerSquaredProcessor] = []
        # The list of all channels that are processed.
        self.channels = []
        # The list of channels that have been processed.
        self.processed_channels = []
        # To prevent user to use ruler or zoom_markers while plotting
        self.is_working = False
        # The post-processing step does not take too much time so there is no
        # need to limit the number of threads that can run at once.
        self.thread_pool = QtCore.QThreadPool()
        self.finished_lock = QtCore.QMutex()

        super().__init__(*args, **kwarg)

    def plot_channels(self, d_obj: DataTypeModel,
                      key: Union[str, Tuple[str, str]],
                      start_tm: float, end_tm: float):
        """
        Recursively plot each TPS channels for waveform_data.
        :param d_obj: object of data
        :param key: data set's key
        :param start_tm: requested start time to read
        :param end_tm: requested end time to read
        """
        self.is_working = True
        self.set_key = key
        self.plotting_data1 = d_obj.waveform_data[key]
        self.plot_total = len(self.plotting_data1)

        self.plotting_bot = const.BOTTOM
        self.plotting_bot_pixel = const.BOTTOM_PX
        self.processed_channels = []
        self.channels = []
        self.tps_processors = []

        start_msg = 'Plotting TPS data...'
        display_tracking_info(self.tracking_box, start_msg)
        self.processing_log = []  # [(message, type)]
        self.gap_bar = None
        self.min_x = max(d_obj.data_time[key][0], start_tm)
        self.max_x = min(d_obj.data_time[key][1], end_tm)

        self.date_mode = self.main_window.date_format.upper()
        if self.plotting_data1 == {}:
            title = "NO WAVEFORM DATA TO DISPLAY TPS."
            self.processing_log.append(
                ("No WAVEFORM data to display TPS.", LogType.INFO))
        else:
            title = get_title(key, self.min_x, self.max_x, self.date_mode)

        self.timestamp_bar_top = self.plotting_axes.add_timestamp_bar(0.)
        self.plotting_axes.set_title(title, y=5, v_align='bottom')

        if self.plotting_data1 == {}:
            self.is_working = False
            self.draw()
            self.clean_up('NO DATA')
            return

        self.start_5mins_of_diff_days = get_start_5mins_of_diff_days(
            self.min_x, self.max_x)
        for chan_id in self.plotting_data1:
            c_data = self.plotting_data1[chan_id]
            if 'tps_data' not in c_data:
                self.channels.append(chan_id)
                channel_processor = TimePowerSquaredProcessor(
                    chan_id, c_data, self.min_x, self.max_x,
                    self.start_5mins_of_diff_days
                )
                channel_processor.signals.finished.connect(self.channel_done)
                channel_processor.signals.stopped.connect(self.channel_done)
                self.tps_processors.append(channel_processor)

        # Because the widget determine if processing is done by comparing the
        # lists of scheduled and finished channels, if a channel runs fast
        # enough that it finishes before any other channel can be scheduled,
        # it will be the only channel executed. To prevent this, we tell the
        # threadpool to only start running the processors once all channels
        # have been scheduled.
        for processor in self.tps_processors:
            self.thread_pool.start(processor)

    @QtCore.Slot()
    def channel_done(self, chan_id: str):
        """
        Slot called when a TPS processor is finished. Plot the TPS data of
        channel chan_id if chan_id is not an empty string and add chan_id to
        the list of processed of channels. If the list of processed channels
        is the same as the list of all channels, notify the user that the
        plotting is finished and add finishing touches to the plot.

        If chan_id is the empty string, notify the user that the plotting has
        been stopped.

        :param chan_id: the name of the channel whose TPS data was processed.
            If the TPS plot is stopped before it is finished, this will be the
            empty string
        """
        self.finished_lock.lock()
        if chan_id != '':
            ax = self.plot_channel(self.plotting_data1[chan_id], chan_id)
            self.axes.append(ax)
        self.processed_channels.append(chan_id)
        if len(self.processed_channels) == len(self.channels):
            self.clean_up(chan_id)
        self.finished_lock.unlock()

    def clean_up(self, chan_id):
        """
        Clean up after all available waveform channels have been stopped or
        plotted. The cleanup steps are as follows.
            Display a finished message
            Add finishing touches to the plot
            Emit the stopped signal of the widget
        """

        if chan_id == '':
            msg = 'TPS plot stopped.'
        else:
            msg = 'TPS plot finished.'
            if chan_id != 'NO DATA':
                self.done()

        display_tracking_info(self.tracking_box, msg)
        self.stopped.emit()

    def done(self):
        """Add finishing touches to the plot and display it on the screen."""
        self.set_legend()
        # Set view size fit with the given data
        if self.main_widget.geometry().height() < self.plotting_bot_pixel:
            self.main_widget.setFixedHeight(self.plotting_bot_pixel)
        self.set_lim_markers()
        self.draw()
        self.is_working = False

    def plot_channel(self, c_data, chan_id):
        """
        TPS is plotted in lines of small rectangular, so called bars.
        Each line is a day so - y value is the order of days
        Each bar is data represent for 5 minutes so x value is the order of
            five minute in a day
        If there is no data in a portion of a day, the bars in the portion
            will have grey color.
        For the five minutes that have data, the color of the bars will be
            based on mapping between tps value of the five minutes against
            the selected color range.

        This function draws each 5 minute with the color corresponding to
        value and create ruler, zoom_marker1, and zoom_marker2 for the channel.

        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to
            DataTypeModel.__init__.waveform_data[key]['read_data'][chan_id]
        :param chan_id: str - name of channel
        :return ax: matplotlib.axes.Axes - axes of the channel
        """

        total_days = c_data['tps_data'].shape[0]
        plot_h = self.plotting_axes.get_height(
            total_days/2, bw_plots_distance=0.003, pixel_height=12.1)
        ax = self.create_axes(self.plotting_bot, plot_h)
        ax.spines[['right', 'left', 'top', 'bottom']].set_visible(False)
        ax.text(
            -0.12, 1,
            f"{get_chan_label(chan_id)} {c_data['samplerate']}sps",
            horizontalalignment='left',
            verticalalignment='top',
            rotation='horizontal',
            transform=ax.transAxes,
            color=self.display_color['plot_label'],
            size=self.font_size + 2
        )

        zoom_marker1 = ax.plot(
            [], [], marker='|', markersize=5,
            markeredgecolor=self.display_color['zoom_marker'])[0]
        self.zoom_marker1s.append(zoom_marker1)

        zoom_marker2 = ax.plot(
            [], [], marker='|', markersize=5,
            markeredgecolor=self.display_color['zoom_marker'])[0]
        self.zoom_marker2s.append(zoom_marker2)

        ruler = ax.plot(
            [], [], marker='s', markersize=4,
            markeredgecolor=self.display_color['time_ruler'],
            markerfacecolor='None')[0]
        self.rulers.append(ruler)

        x = np.array([i for i in range(const.NO_5M_DAY)])
        square_counts = self.parent.sel_square_counts  # square counts range
        color_codes = self.parent.color_def  # colordef

        # --------------------------- PLOT TPS -----------------------------#
        for dayIdx, y in enumerate(c_data['tps_data']):
            # not draw data out of day range
            color_set = self.get_color_set(y, square_counts, color_codes)
            # (- dayIdx): each day is a line, increase from top to bottom
            ax.scatter(x, [- dayIdx] * len(x), marker='s',
                       c=color_set, s=3)
        # extra to show highlight square
        ax.set_ylim(-(c_data['tps_data'].shape[0] + 1), 1)

        return ax

    def set_legend(self):
        """
        Plot one dot for each color and assign label to it. The dots are
            plotted outside of xlim to not show up in plotting area. xlim is
            set so that it has some extra space to show full hightlight square
            of the ruler.
        ax.legend will create one label for each dot.
        """
        # set height of legend and distance bw legend and upper ax
        plot_h = self.plotting_axes.get_height(
            21, bw_plots_distance=0.004, pixel_height=12)
        ax = self.plotting_axes.canvas.figure.add_axes(
            [self.plotting_l, self.plotting_bot, self.plotting_w, plot_h],
            picker=True
        )
        ax.axis('off')
        ax.patch.set_alpha(0)
        c_labels = self.parent.sel_col_labels
        clrs = self.parent.color_def  # colordef
        for idx in range(len(c_labels)):
            # draw a dot out of xlim so it isn't displayed in plotting area
            ax.scatter([300], [1],
                       c=clr[clrs[idx]],
                       s=0.1,
                       label=c_labels[idx],
                       edgecolor=self.display_color['basic'],
                       alpha=0.8,
                       zorder=1,
                       picker=True)
        # extra to show highlight square
        ax.set_xlim(-2, const.NO_5M_DAY + 1)
        ax.legend(loc="upper left", framealpha=0.2,
                  markerscale=25,
                  labelcolor=self.display_color['basic'])

    def get_color_set(self, y, square_counts, colors):
        """
        Create array of color (col) according to value of y compare with
            square_counts (square count range)

        :param y: np.array of float - tps values for all 5-minutes of a day
        :param square_counts: [int, ] - list of square count ranges
        :param colors: [str,] - list of color codes based on colorSettings.clr
        :return: np.array of color hex for each 5-minutes of a day
        """
        return (
            np.where(
                y == square_counts[0], clr[colors[0]], np.where(
                    y < square_counts[1], clr[colors[1]], np.where(
                        y < square_counts[2], clr[colors[2]], np.where(
                            y < square_counts[3], clr[colors[3]], np.where(
                                y < square_counts[4], clr[colors[4]], np.where(
                                    y < square_counts[5], clr[colors[5]], np.where(             # noqa: E501
                                        y < square_counts[6], clr[colors[6]], clr[colors[7]]    # noqa: E501
                                    )))))))
        )

    def create_axes(self, plot_b, plot_h):
        """
        Create axes for 288 of 5m in a day in which minor tick for every hour,
            major tick for every 4 hour

        :param plot_b: float - bottom of the plot
        :param plot_h: float - height of the plot
        :return ax: matplotlib.axes.Axes - axes of tps of a waveform channel
        """
        ax = self.plotting_axes.canvas.figure.add_axes(
            [self.plotting_l, plot_b, self.plotting_w, plot_h],
            picker=True
        )
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.xaxis.grid(True, which='major',
                      color=self.display_color['basic'], linestyle='-')
        ax.xaxis.grid(True, which='minor',
                      color=self.display_color['sub_basic'], linestyle='-')
        ax.set_yticks([])

        times, major_times, major_time_labels = get_day_ticks()
        ax.set_xticks(times, minor=True)
        ax.set_xticks(major_times)
        ax.set_xticklabels(major_time_labels, fontsize=self.font_size,
                           color=self.display_color['basic'])
        # extra to show highlight square
        ax.set_xlim(-2, const.NO_5M_DAY + 1)
        ax.patch.set_alpha(0)
        return ax

    def on_pick_event(self, event):
        """
        When a plot is select, the corresponding point on each plot will
            be highlighted and their time and counts will be displayed.
        :param event: pick event - event when object of canvas is selected.
            The event happens before button_press_event.
        """
        if event.mouseevent.name == 'scroll_event':
            return
        if event.mouseevent.button in ('up', 'down'):
            return
        info_str = ""
        if event.artist in self.axes:
            xdata = event.mouseevent.xdata
            if xdata is None:
                return
            xdata = round(xdata)                        # x value on the plot
            # when click on outside xrange that close to edge, adjust to edge
            if xdata in [-2, -1]:
                xdata = 0
            if xdata in [288, 289]:
                xdata = 287
            ydata = round(event.mouseevent.ydata)       # y value on the plot

            # refer to description in plot_channel to understand x,y vs
            # day_index, five_min_index
            day_index = - ydata
            five_min_index = xdata
            try:
                # identify time for rulers on other plotting widget
                self.tps_t = self.start_5mins_of_diff_days[
                    day_index, five_min_index]
                format_t = format_time(self.tps_t, self.date_mode, 'HH:MM:SS')
                info_str += f"<pre>{format_t}:"
                for chan_id in self.plotting_data1:
                    c_data = self.plotting_data1[chan_id]
                    data = c_data['tps_data'][day_index, five_min_index]
                    info_str += (f"   {chan_id}:"
                                 f"{add_thousand_separator(sqrt(data))}")
                info_str += "  (counts)</pre>"
                display_tracking_info(self.tracking_box, info_str)
                self.draw()
            except IndexError:
                # exclude the extra points added to the 2 sides of x axis to
                # show the entire highlight box
                pass

    def on_ctrl_cmd_click(self, xdata):
        """
        Ctrl + cmd: base on xdata to find indexes of x an y
            in self.each_day_5_min_list to display ruler for each channel, then
            set time for each ruler in self.rulers to place them in correct
            position.

        :param xdata: float - time value in other plot
        """
        self.zoom_marker1_shown = False
        x_idx, y_idx = find_tps_tm_idx(xdata, self.start_5mins_of_diff_days)
        for rl in self.rulers:
            rl.set_data(x_idx, y_idx)

    def on_shift_click(self, xdata):
        """
        Shift + right click on other plot widget, this function will be called
            to show marker for place when it is zoomed.
            On the fist of zoom_marker, make ruler disappeared, set min_x.
            On the second of zoom_maker, call set_lim_markers to mark the new
                limit.

        :param xdata: float - time value in other plot
        """

        if not self.zoom_marker1_shown:
            self.set_rulers_invisible()
            self.min_x = xdata
            self.zoom_marker1_shown = True
        else:
            [self.min_x, self.max_x] = sorted(
                [self.min_x, xdata])
            self.set_lim_markers()
            self.zoom_marker1_shown = False

    def set_rulers_invisible(self):
        """
        Clear data for self.rulers to make them disappeared.
        """
        for rl in self.rulers:
            rl.set_data([], [])

    def set_lim_markers(self):
        """
        Find x index (which index in five minutes of a day) and
            y index (which day) of self.min_x and self.min_y, and set data for
            all markers in self.zoom_marker1s and self.zoom_marker2s.
        """
        x_idx, y_idx = find_tps_tm_idx(self.min_x,
                                       self.start_5mins_of_diff_days)
        for zm1 in self.zoom_marker1s:
            zm1.set_data(x_idx, y_idx)
        x_idx, y_idx = find_tps_tm_idx(self.max_x,
                                       self.start_5mins_of_diff_days)
        for zm2 in self.zoom_marker2s:
            zm2.set_data(x_idx, y_idx)

    def request_stop(self):
        """Request all running channel processors to stop."""
        for processor in self.tps_processors:
            processor.request_stop()

    def replot(self):
        """
        Reuse tps_data calculated in the first plotting to replot
        with new color range selected.
        """
        self.clear()
        self.plotting_bot = const.BOTTOM
        title = get_title(self.set_key, self.min_x, self.max_x, self.date_mode)
        self.timestamp_bar_top = self.plotting_axes.add_timestamp_bar(0.)
        self.plotting_axes.set_title(title, y=0, v_align='bottom')
        for chan_id in self.plotting_data1:
            c_data = self.plotting_data1[chan_id]
            self.plot_channel(c_data, chan_id)
        self.done()


class TimePowerSquaredDialog(QtWidgets.QWidget):
    def __init__(self, parent):
        """
        Dialog to display time power square data for waveform channels. Users
            are allowed to select different Color Ranges (Antarctica, Low,
            Medium and High) for differences values vs. colors map to draw
            time power square for each 5-minute data

        :param parent: QMainWindow/QWidget - the parent widget
        """
        super().__init__()
        self.parent = parent
        """
        data_type: str - type of data being plotted
        """
        self.data_type = None
        self.setGeometry(50, 50, 1200, 700)
        self.setWindowTitle("TPS Plot")

        main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(main_layout)
        main_layout.setContentsMargins(5, 5, 5, 5)
        main_layout.setSpacing(0)

        """
        tracking_info_text_browser: QTextBrowser - to display info text
        """
        self.info_text_browser = QtWidgets.QTextBrowser(self)
        """
        plotting_widget: PlottingWidget - the widget to draw time-power-square
            for each 5-minute of data
        """
        self.plotting_widget = TimePowerSquaredWidget(
            self, self.info_text_browser, "TPS", self.parent)
        main_layout.addWidget(self.plotting_widget, 2)

        bottom_layout = QtWidgets.QHBoxLayout()
        bottom_layout.addSpacing(20)
        main_layout.addLayout(bottom_layout)

        buttons_layout = QtWidgets.QVBoxLayout()
        bottom_layout.addLayout(buttons_layout)

        # ################ Color range #################
        color_layout = QtWidgets.QHBoxLayout()
        buttons_layout.addLayout(color_layout)
        color_layout.addWidget(QtWidgets.QLabel("Color Range"))

        """
        color_def: [str,] -  list of color codes in order of values to be
            displayed
        """
        self.color_def = get_color_def()
        """
        sel_square_counts: [int,] - selected time-power-square ranges
        """
        self.sel_square_counts = []
        """
        sel_col_labels: [str,] - color label for each range
        """
        self.sel_col_labels = []
        """
        color_ranges: [str,] - name of color:value map for user to choose
            'Antarctica'/'Low'/'Medium'/'High'
        all_square_counts: [[int,],] - time-power-squared values
        color_label: [str,] - labels that define count range for colors to show
            in the legend of the tps plotting widget
        """
        (self.color_ranges,
         self.all_square_counts,
         self.color_label) = get_color_ranges()

        """
        color_range_choice: QComboBox - dropdown box for user to choose a
            color:value map name
        """
        self.color_range_choice = QtWidgets.QComboBox(self)
        self.color_range_choice.addItems(self.color_ranges)
        self.color_range_choice.setCurrentText('High')
        color_layout.addWidget(self.color_range_choice)
        # ##################### Replot button ########################
        self.replot_button = QtWidgets.QPushButton("RePlot", self)
        buttons_layout.addWidget(self.replot_button)

        # ##################### Save button ##########################
        self.save_button = QtWidgets.QPushButton('Save', self)
        buttons_layout.addWidget(self.save_button)

        self.info_text_browser.setFixedHeight(60)
        bottom_layout.addWidget(self.info_text_browser)

        self.connect_signals()
        self.color_range_changed()

    def set_data(self, data_type, file_name):
        """
        Set data_type and the window's title.

        :param data_type: str - data type of data being plotted
        :param file_name: str - name of the file/folder of the data set to be
            displayed
        """
        self.data_type = data_type
        self.setWindowTitle("TPS Plot %s - %s" % (data_type, file_name))

    def resizeEvent(self, event):
        """
        OVERRIDE Qt method.
        When TimePowerDialog is resized, its plotting_widget need to initialize
            its size to fit the viewport.

        :param event: QResizeEvent - resize event
        """
        self.plotting_widget.init_size()

    def connect_signals(self):
        """
        Connect functions to widgets
        """
        self.save_button.clicked.connect(self.save)
        self.replot_button.clicked.connect(self.plotting_widget.replot)
        self.color_range_choice.currentTextChanged.connect(
            self.color_range_changed)

    @QtCore.Slot()
    def color_range_changed(self):
        """
        When selecting a different color range from the drop down menu, set
            new value for sel_square_counts and sel_col_labels
        """
        color_range = self.color_range_choice.currentText()
        cr_index = self.color_ranges.index(color_range)
        self.sel_square_counts = self.all_square_counts[cr_index]
        self.sel_col_labels = self.color_label[cr_index]

    @QtCore.Slot()
    def save(self):
        """
        Save the plotting to a file
        """
        print("save")
